#define _DEFAULT_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <stdint.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sched.h>

#include <timer.h>

#include <omp.h>

#include <likwid.h>

#ifndef ROUNDS
#define ROUNDS 100
#endif

#ifndef SIZE
#define SIZE 4000000
#endif

#define gettid() syscall(SYS_gettid)
#define HLINE "-------------------------------------------------------------\n"

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

void copy(int size, double* dst, double* src);
void scale(int size, double* dst, double* src, double scalar);
void add(int size, double* dst, double* src1, double* src2);
void triad(int size, double* dst, double* src1, double* src2, double* src3);


void __attribute__ ((noinline)) allocate_vector(double** ptr, uint64_t size, double value)
{
    int j = 0;
    int errorCode = 0;

    errorCode = posix_memalign((void**) ptr, 64, size*sizeof(double));

    if (errorCode)
    {
        if (errorCode == EINVAL)
        {
            fprintf(stderr,
                "Alignment parameter is not a power of two\n");
            exit(EXIT_FAILURE);
        }
        if (errorCode == ENOMEM)
        {
            fprintf(stderr,
                "Insufficient memory to fulfill the request\n");
            exit(EXIT_FAILURE);
        }
    }
    for (j=0; j<SIZE; j++) {
        (*ptr)[j] = value;
    }
}

void __attribute__ ((noinline)) print_result(char* name, int streams, double time)
{
    printf("Processed %.1f Mbyte at %s benchmark in %.4f seconds: %.2f MByte/s\n",
                        1E-6*(streams*SIZE*sizeof(double)),
                        name, time,
                        1E-6*((streams*SIZE*sizeof(double))/time));
}


int main(int argn, char** argc)
{

    int j = 0, k = 0;
    double *a,*b,*c,*d;
    double scalar = 3.0;
    TimeData timer;
    double triad_time, copy_time, scale_time, add_time;
    LIKWID_MARKER_INIT;
#pragma omp parallel
{
    LIKWID_MARKER_REGISTER("copy");
    LIKWID_MARKER_REGISTER("scale");
    LIKWID_MARKER_REGISTER("add");
    LIKWID_MARKER_REGISTER("triad");
}

    allocate_vector(&a, SIZE, 1.0);
    allocate_vector(&b, SIZE, 2.0);
    allocate_vector(&c, SIZE, 3.0);
    allocate_vector(&d, SIZE, 2.0);


    time_start(&timer);
    {
        for (k=0; k<ROUNDS; k++)
        {
            copy(SIZE, c, a);
        }
    }
    time_stop(&timer);
    copy_time = time_print(&timer)/(double)ROUNDS;

    time_start(&timer);
    {
        for (k=0; k<ROUNDS; k++)
        {
            scale(SIZE, b, c, scalar);
        }
    }
    time_stop(&timer);
    scale_time = time_print(&timer)/(double)ROUNDS;

    time_start(&timer);
    {
        for (k=0; k<ROUNDS; k++)
        {
            add(SIZE, c, a, b);
        }
    }
    time_stop(&timer);
    add_time = time_print(&timer)/(double)ROUNDS;

    time_start(&timer);
    {
        for (k=0; k<ROUNDS; k++)
        {
            triad(SIZE, a, b, c, d);
        }
    }
    time_stop(&timer);
    triad_time = time_print(&timer)/(double)ROUNDS;

    print_result("copy", 2, copy_time);
    print_result("scale", 2, scale_time);
    print_result("add", 3, add_time);
    print_result("triad", 4, triad_time);


    free(a);
    free(b);
    free(c);
    free(d);
    LIKWID_MARKER_CLOSE;
    return 0;
}

