/**
 *       @file  timer.h
 *      @brief  Measure runtime with getTimeOfday and rdtsc.
 *
 * A C module to measure runtime. There are two methods: with gettimeofday for
 * longer time periods (over 0.5 sec) and with rdtsc (read time stamp counter)
 * for shorter periods. There is a variation for measurements with rdtsc
 * of 100 cycles in the worst case. Therefore sensible measurements should be
 * over 1000 cycles.
 *
 *     @author  Jan Treibig (jt), jan.treibig@gmail.com
 *
 *   @internal
 *     Created  07/30/2009
 *    Revision  ---
 *    Compiler  gcc
 *     Company  RRZE Erlangen
 *   Copyright  Copyright (c) 2009, Jan Treibig
 *
 * This source code is released for free distribution under the terms of the
 * GNU General Public License as published by the Free Software Foundation.
 * ============================================================================
 */
#ifndef TIMER_H
#define TIMER_H

#include "timer_types.h"

#define RDTSC2(cpu_c) \
__asm__ volatile( "rdtsc\n\t"           \
"movl %%eax, %0\n\t"  \
"movl %%edx, %1\n\t"  \
: "=r" ((cpu_c).int32.lo), "=r" ((cpu_c).int32.hi) \
: : "%eax", "%edx")

#define RDTSC(cpu_c) \
__asm__ volatile("xor %%eax,%%eax\n\t"           \
"cpuid\n\t"           \
"rdtsc\n\t"           \
"movl %%eax, %0\n\t"  \
"movl %%edx, %1\n\t"  \
: "=r" ((cpu_c).int32.lo), "=r" ((cpu_c).int32.hi) \
: : "%eax","%ebx","%ecx","%edx")



/**
 * @brief  Initialize timer module
 *
 * Determines processor clock and cycles for cpuid.
 */
extern void time_init(void );

/**
 * @brief  Start timer measurement with getTimeofDay
 * @param  time  Reference to struct holding the timestamps
 */
extern void time_start(TimeData* time);

/**
 * @brief  Stop timer measurement with getTimeofDay
 * @param  time Reference to struct holding the timestamps
 */
extern void time_stop(TimeData* time);

/**
 * @brief  Get timer measurement with getTimeofDay
 * @param  time Reference to struct holding the timestamps
 * @return Time duration between start and stop in seconds
 */
extern float time_print(TimeData* timer);

/**
 * @brief  Start cycles measurement with rdtsc
 * @param cycles Reference to struct holding the timestamps
 */
extern void time_startCycles(CyclesData* cycles);

/**
 * @brief  Stop cycles measurement with rdtsc
 * @param cycles Reference to struct holding the timestamps
 */
extern void time_stopCycles(CyclesData* cycles);


/**
 * @brief  Get time of cycles measurement
 * @param cycles Reference to struct holding the timestamps
 * @return Timer duration between start and stop in seconds
 */
extern float time_printCyclesTime(CyclesData* cycles);

/**
 * @brief  Get cycles of cycles measurement
 * @param cycles Reference to struct holding the timestamps
 * @return cycles between start and stop
 */
extern uint64_t time_printCycles(CyclesData* cycles);

/**
 * @brief  Get Clock rate of cpu in Hertz
 * @return clock rate of cpu
 */
extern uint64_t time_getCpuClock(void);

/**
 * @brief  Get cycles for cpuid
 * @return cycles for cpuid
 */
extern uint64_t time_getCpuidCycles(void);

#endif /* TIMER_H */
